/*
* This program and the accompanying materials are made available under the terms of the
* Eclipse Public License v2.0 which accompanies this distribution, and is available at
* https://www.eclipse.org/legal/epl-v20.html
*
* SPDX-License-Identifier: EPL-2.0
*
* Copyright Contributors to the Zowe Project.
*
*/

import { ICommandDefinition } from "../../src/doc/ICommandDefinition";
import { CommandPreparer } from "../../src/CommandPreparer";
import { ICommandOptionDefinition } from "../../src/doc/option/ICommandOptionDefinition";
import { COMPLEX_COMMAND, MULTIPLE_GROUPS } from "../__resources__/CommandDefinitions";
import { TestLogger } from "../../../../__tests__/TestLogger";
import { inspect } from "util";
import { UnitTestUtils } from "../../../../__tests__/src/UnitTestUtils";
import { CommandUtils, ICommandTreeEntry } from "../../";
UnitTestUtils.replaceIt();

describe("Command Utils", () => {
    it("We should be able to determine if an option is specified", () => {
        expect(CommandUtils.optionWasSpecified("test-option", {
            name: "test-command",
            description: "my command",
            type: "command",
            options: [
                {
                    name: "test-option",
                    description: "the option",
                    type: "string"
                }
            ]
        }, {"$0": undefined, "_": undefined, "test-option": "This was specified"})).toEqual(true);
    });

    it("We should be able to build the full command string from the definition and the arguments", () => {
        const expected: string = "maincommand test-group test-command \"value2\" --test-option value1 --test-boolean";
        const preparedCommand: ICommandDefinition = CommandPreparer.prepare(COMPLEX_COMMAND.children[0]);
        const constructed: string = CommandUtils.reconstructCommand({
            "$0": undefined,
            "_": ["test-group", "test-command"],
            "test-option": "value1",
            "positional1": "value2",
            "test-boolean": true
        }, preparedCommand);
        expect(constructed).toEqual(expected);
    });

    it("We should be able to get the option definition from the name", () => {
        const optionName: string = "test-option";
        const option: ICommandOptionDefinition = {
            name: optionName,
            description: "the option",
            type: "string"
        };
        expect(CommandUtils.getOptionDefinitionFromName(optionName, {
            name: "test-command",
            description: "my command",
            type: "command",
            options: [option]
        })).toEqual(option);
    });

    it("We should be able to flatten a nested command tree for display and searching purposes", () => {
        const flatten: ICommandTreeEntry[] = CommandUtils.flattenCommandTree(COMPLEX_COMMAND);
        TestLogger.info("Flattened Command Tree:\n" + inspect(flatten));
        expect(flatten).toMatchSnapshot();
    });

    it("We should be able get the full command name from the flattened tree", () => {
        const fullCommand: string = CommandUtils.getFullCommandName(COMPLEX_COMMAND.children[0], MULTIPLE_GROUPS);
        expect(fullCommand).toMatchSnapshot();
    });
});
